from __future__ import annotations
from itertools import combinations
from typing import Iterable, Union, Tuple
from .axis import AxisPoint

# This is a hack to make mypy happy
if False:
    from .matrix import RiskMatrix


class Coordinate:
    """A collection of AxisPoints to represent a location in a matrix."""

    def __init__(self, points: Iterable[AxisPoint]) -> None:
        if any(p.axis is None for p in points):
            raise ValueError(
                "There is at least one point which is not tied to an axis."
            )

        if any(p.axis.matrix is None for p in points):
            raise ValueError(
                "There is at least one point with an axis that is not tied to a matrix."
            )

        point_pairs = (pair for pair in combinations(points, 2))
        pair_sets = (set((pair[0].axis, pair[1].axis)) for pair in point_pairs)
        if any((len(s) == 1 for s in pair_sets)):
            raise ValueError(
                "There are two points on the same axis. Every point of a coordinate should be on a different axis."
            )

        if len(set(p.axis.matrix for p in points)) != 1:
            raise ValueError(
                "The points in this coordinate are not all from the same matrix."
            )

        self.matrix: RiskMatrix = set(p.axis.matrix for p in points).pop()
        # Order the points in the coordinate with the same order as the axes in the matrix.
        _points = []
        for axis in self.matrix.axes:
            for point in points:
                if axis == point.axis:
                    _points.append(point)
        self.points: Tuple[AxisPoint, ...] = tuple(_points)

    def __repr__(self):
        return f"Coordinate({[p.code for p in self.points]})"

    def __str__(self) -> str:
        """ The str format is also used for determining equality with coordinates in string format """
        return f'{"".join(str(p.code) for p in self.points)}'

    def __getitem__(self, key: int) -> AxisPoint:
        return self.points[key]

    def __hash__(self) -> int:
        return hash(self.points)

    def __eq__(self, other: object) -> bool:
        """Check if two Coordinates are pointing to the same location.

        Args:
            other (Union[Coordinate, str]): A Coordinate object or string version.

        Returns:
            bool: Return True if coordinate location is equal, return False otherwise.
        """
        if not isinstance(other, (Coordinate, str)):
            return NotImplemented
        if isinstance(other, str):
            return str(self) == other
        return self.points == other.points

    def __lt__(self, other: object) -> bool:
        if not isinstance(other, Coordinate):
            return NotImplemented

        value_self = sum(p.value for p in self.points)
        value_other = sum(p.value for p in other.points)

        # If we need an order even if two coordinates have equivalent values,
        # compare the strings. This means the coordinates are ordered alphabetically
        # which is determined by the order of the axes.
        if value_self == value_other and self.matrix.force_coordinate_order:
            return str(self) < str(other)

        return value_self < value_other

    @property
    def value(self) -> int:
        return sum(p.value for p in self.points)

    @property
    def category(self):
        return self.matrix.get_category(self)



import pytest
from riskmatrix import RiskMatrix, Axis, AxisPoint, Category, Coordinate, Point


@pytest.fixture
def rm():
    return RiskMatrix("Risk Matrix name")


@pytest.fixture
def rm_points():
    return (
        Point("A", "Unlikely"),
        Point("B", "Likely"),
        Point("C", "Very Likely"),
        Point(1, "No Impact"),
        Point(2, "Cheap"),
        Point(3, "Expensive"),
    )


@pytest.fixture
def rm_categories():
    return (
        ("LOW", "Low risk", "#ffff11", "#ffffff"),
        ("MED", "Med risk", "#ffff00", "#ffffff"),
        ("HIG", "Hig risk", "#ff0000", "#ffffff"),
    )


@pytest.fixture
def rm_with_axis(rm, rm_points):
    a, b, c, one, two, three = rm_points

    rm.add_axis("x", points=[a, b, c])
    rm.add_axis("y", points=[one, two, three])

    return rm


@pytest.fixture
def rm_with_categories(rm_with_axis, rm_categories):
    rm = rm_with_axis
    low, med, hig = rm_categories

    rm.add_category(*low)
    rm.add_category(*med)
    rm.add_category(*hig)

    return rm


@pytest.fixture
def rm_coordinates(rm_with_categories):
    rm = rm_with_categories
    a, b, c = rm.axes["x"].points
    one, two, three = rm.axes["y"].points

    return (
        Coordinate((a, one)),
        Coordinate((a, two)),
        Coordinate((a, three)),
        Coordinate((b, one)),
        Coordinate((b, two)),
        Coordinate((b, three)),
        Coordinate((c, one)),
        Coordinate((c, two)),
        Coordinate((c, three)),
    )


@pytest.fixture
def rm_full(rm_with_categories, rm_coordinates):
    rm = rm_with_categories
    low, med, hig = rm.categories

    a1, a2, a3, b1, b2, b3, c1, c2, c3 = rm_coordinates

    rm.map_coordinates(low, [a1, a2, a3, b1])
    rm.map_coordinates(med, [b2, c1])
    rm.map_coordinates(hig, [b3, c2, c3])

    return rm


class TestRiskMatrix:
    def test_riskmatrix_name(self):
        name = "Test name 1"
        rm = RiskMatrix(name)

        assert rm.name == name

        name2 = "Test name 2"

        assert rm.name != name2

        rm.name = name2

        assert rm.name == name2
        assert rm.name != name

    def test_sorted_coordinates(self, rm_full):
        rm = rm_full
        raw_data = ["A1", "B2", "C3", "A3", "B1", "B1", "A2"]

        # Get top 2 values
        coordinates = (rm.get_coordinate(c) for c in raw_data)
        r = sorted(coordinates, reverse=True)[:2]

        assert r == [rm.get_coordinate("C3"), rm.get_coordinate("B2")]

    def test_coordinates_have_right_category(self, rm_full):
        # Each category has all coordinates
        # These are the categories in rm_full:
        # low =  a1, a2, a3, b1
        # med =  b2, c1
        # hig =  b3, c2, c3
        rm = rm_full

        a1, a2, a3, b1, b2, b3, c1, c2, c3 = rm.coordinates
        low, med, hig = rm.categories

        assert a1.category == low
        assert a2.category == low
        assert a3.category == low
        assert b1.category == low
        assert b2.category == med
        assert c1.category == med
        assert b3.category == hig
        assert c2.category == hig
        assert c3.category == hig

    def test_coordinates_order_alphabetical(self, rm_full, rm_coordinates):
        rm = rm_full
        a1, a2, a3, b1, b2, b3, c1, c2, c3 = rm_coordinates
        a1a, a2a, a3a, b1a, b2a, b3a, c1a, c2a, c3a = rm.coordinates

        assert a1 == a1a
        assert a2 == a2a
        assert a3 == a3a
        assert b1 == b1a
        assert b2 == b2a
        assert b3 == b3a
        assert c1 == c1a
        assert c2 == c2a
        assert c3 == c3a

    def test_coordinate_median(self, rm_full):
        rm = rm_full
        raw_data = ["A1", "B2", "C3", "A3", "B1", "B1", "A2"]

        # Convert into coordinates
        coordinates = (rm.get_coordinate(c) for c in raw_data)

        # Calculate median
        import statistics

        median = statistics.median(coordinates)

        assert median == rm_full.get_coordinate("B1")


class TestAxis:
    def test_add_axis_with_points(self, rm, rm_points):
        a, b, c, *_ = rm_points

        rm.add_axis("x", points=[a, b, c])
        aa, bb, cc = rm.axes["x"].points

        assert a.code == aa.code
        assert a.name == aa.name
        assert a.description == aa.desc
        assert a.code != bb.code
        assert b.code == bb.code
        assert rm.axes["x"][0] is aa
        assert rm.axes["x"][1] is bb
        assert rm.axes["x"][2] is cc

    def test_add_axis_conflicting_named_arguments(self, rm, rm_points):
        a, b, c, *_ = rm_points

        try:
            rm.add_axis("x", points=[a, b, c], size=4)
        except ValueError as e:
            assert (
                str(e)
                == "You should choose between giving a list of points or defining a size."
            )

    def test_add_axis_requires_named_arguments(self, rm, rm_points):
        a, b, c, *_ = rm_points

        try:
            rm.add_axis("x", [a, b, c])
        except TypeError as e:
            assert str(e).startswith("add_axis() takes 2 positional arguments")

    def test_add_axis_with_size(self, rm):
        rm.add_axis("x", size=4)

        assert rm.axes["x"][0].code == "1"
        assert rm.axes["x"][1].code == "2"
        assert rm.axes["x"][2].code == "3"
        assert rm.axes["x"][3].code == "4"
        assert len(rm.axes["x"]) == 4

    def test_convert_number_to_letter(self, rm):
        assert rm._convert_number_to_letter(1) == "A"
        assert rm._convert_number_to_letter(26) == "Z"

        try:
            number = 27
            rm._convert_number_to_letter(number)
        except ValueError as e:
            assert str(e) == f"The number {number} has to be between 1 and 26."

        try:
            number = 0
            rm._convert_number_to_letter(number)
        except ValueError as e:
            assert str(e) == f"The number {number} has to be between 1 and 26."


class TestCategory:
    def test_add_category(self, rm, rm_categories):
        low, med, hig = rm_categories

        rm.add_category(*low)
        rm.add_category(*med)
        rm.add_category(*hig)

        assert len(rm.categories) == 3

        assert rm.categories[0].value == 0
        assert rm.categories[1].value == 1
        assert rm.categories[2].value == 2

    def test_get_categories(self, rm_with_categories, rm_categories):
        low_origin, med_origin, hig_origin = rm_categories
        low, med, hig = rm_with_categories.categories

        assert (low.code, low.name, low.color, low.text_color) == (low_origin)
        assert (med.code, med.name, med.color, med.text_color) == (med_origin)
        assert (hig.code, hig.name, hig.color, hig.text_color) == (hig_origin)

    def test_category_value(self, rm_with_categories):
        low, med, hig = rm_with_categories.categories

        assert low.value == 0
        assert med.value == 1
        assert hig.value == 2

    def test_max_category(self, rm_full):
        assert rm_full.get_max_category().value == 2


class TestCoordinate:
    def test_add_coordinate(self, rm_with_categories, rm_coordinates):
        rm = rm_with_categories

        low, med, hig = rm.categories
        a1, a2, a3, b1, b2, b3, c1, c2, c3 = rm_coordinates

        rm.map_coordinates(low, [a1, a2, a3, b1])
        rm.map_coordinates(med, (b2, b3, c1, c2))
        rm.map_coordinate(hig, c3)

        assert rm.get_category(a3) == low
        # assert rm.get_category(b2) == med
        assert rm.get_category(c3) == hig

    def test_add_multiple_points_on_one_axis(self, rm_with_axis):
        rm = rm_with_axis

        a, b, _ = rm.axes["x"].points

        try:
            Coordinate((a, b))
        except ValueError as e:
            assert (
                str(e)
                == "There are two points on the same axis. Every point of a coordinate should be on a different axis."
            )
        else:
            assert False

    def test_get_coordinate(self, rm_with_categories):
        rm = rm_with_categories
        low, med, hig = rm.categories
        a, b, c = rm.axes["x"].points
        one, two, three = rm.axes["y"].points

        rm.map_coordinates(low, [(a, one), (a, two), (a, three), (b, one)])

        rm.map_coordinates(med, [(b, two), (b, three), (c, one), (c, two)])

        c3 = rm.map_coordinate(hig, (c, three))

        get_c3 = rm.get_coordinate("C3")
        assert get_c3 == c3
        assert rm.get_category(get_c3) == hig

    def test_equivalent_coordinate_value(self, rm_full):
        """Test if we can force a coordinate order if two Coordinates have
        an equal value by using alphabetical order. Then also test if we can
        turn it off and let max() return the first of the equal values.
        """
        rm = rm_full

        c2 = rm.get_coordinate("C2")
        b3 = rm.get_coordinate("B3")

        rm.force_coordinate_order = True
        assert max(b3, c2) == c2

        rm.force_coordinate_order = False
        assert max(b3, c2) == b3

    def test_coordinate_code(self, rm_with_categories):

        rm = rm_with_categories
        a, *_ = rm.axes["x"].points
        one, *_ = rm.axes["y"].points

        a1 = Coordinate((a, one))
        a1_reverse = Coordinate((one, a))

        assert str(a1) == "A1"
        assert str(a1) != "1A"
        assert str(a1_reverse) == "A1"
        assert str(a1_reverse) != "1A"

    def test_max_coordinate(self, rm_full):
        rm = rm_full
        max_coordinate = rm.coordinates[-1]

        assert max(rm.coordinates) == max_coordinate
        # This should work once there is an unambiguous way to determine what the maximum coordinate is
        # assert max(rm.coordinates) == max_coordinate

    def test_add_wrong_coordinate(self, rm_with_categories):
        pass



from __future__ import annotations
from typing import Optional
from functools import total_ordering


@total_ordering
class Category:
    """A collection of points has a RiskMatrix Category."""

    def __init__(
        self, code: str, name: str, color: str, text_color: str, description: str = ""
    ) -> None:
        self.code = code
        self.name = name
        self.desc = description
        self.color = color
        self.text_color = text_color
        self.value: int = 0

    def __repr__(self):
        return f"Category{self.code, self.name, self.desc}"

    def __str__(self):
        return f"Category: {self.code} - {self.name}"

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Category):
            return NotImplemented
        return self.value == other.value

    def __lt__(self, other: object) -> bool:
        if not isinstance(other, Category):
            return NotImplemented
        return self.value < other.value



"""
==========
Riskmatrix
==========

A library to deal with risk matrix data.
"""
__version__ = "0.0.9"

from .axis import Axis, AxisPoint, Point
from .category import Category
from .coordinate import Coordinate
from .matrix import RiskMatrix



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages

with open("README.rst") as readme_file:
    readme = readme_file.read()

with open("HISTORY.rst") as history_file:
    history = history_file.read()

requirements = []  # type: ignore

setup_requirements = ["pytest-runner"]

test_requirements = ["pytest"]

setup(
    author="Alex de Ruijter",
    author_email="alex@slicerisk.com",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
    ],
    description="A package to deal with risk matrix data.",
    install_requires=requirements,
    license="GNU General Public License v3",
    long_description=readme + "\n\n" + history,
    include_package_data=True,
    keywords="riskmatrix",
    name="riskmatrix",
    packages=find_packages(include=["riskmatrix"]),
    setup_requires=setup_requirements,
    test_suite="tests",
    tests_require=test_requirements,
    url="https://github.com/slicerisk/riskmatrix",
    version="0.0.9",
    zip_safe=False,
)



from typing import Dict, Iterable, Optional, Tuple, Union
from . import Axis, AxisPoint, Category, Coordinate, Point


class AxesDescriptor:
    """A descriptor for the axes attribute on the RiskMatrix.

    Raises:
        KeyError: When an axis name is not found in the axes.
        AttributeError: Overriding or deleting the axes attribute is not allowed.
    """

    def __init__(self):
        self._axes = []

    def __get__(self, instance, owner) -> Tuple[Axis, ...]:
        return tuple(self._axes)

    def __getitem__(self, val: Union[str, int]) -> Axis:
        if isinstance(val, str):
            for axis in self._axes:
                if axis.name == val:
                    return axis
            raise KeyError(f"No axis named {val}.")

        if isinstance(val, int):
            return self._axes[val]

        raise TypeError(f"Axes indices must be integers, or strings, not {val}")

    def __iter__(self):
        self.__i = 0
        return self

    def __next__(self):
        try:
            val = self[self.__i]
        except IndexError:
            raise StopIteration
        self.__i += 1
        return val

    def __set__(self, instance, val):
        raise AttributeError("Can't override axes attribute. Add axes individually.")

    def __delete__(self, instance):
        raise AttributeError("Can't delete axes attribute.")

    def add(self, axis: Axis):
        self._axes.append(axis)


class RiskMatrix:
    """The main class to build a risk matrix."""

    def __init__(self, name: str) -> None:
        self.name = name
        self.axes: AxesDescriptor = AxesDescriptor()
        self._categories: Dict[int, Category] = {}
        self._coordinates: Dict[Coordinate, Category] = {}

        # This boolean determines whether it's ok to force Coordinate order if they have a similar value.
        # Setting it to False can make ordering Coordinates with equivalent values ambiguous.
        self.force_coordinate_order = True

    def __repr__(self):
        return f"RiskMatrix({self.name}) " + str(self.axes)

    def __str__(self):
        return self.name

    @property
    def categories(self) -> Tuple[Category, ...]:
        """Return a tuple of all Categories in the Riskmatrix.

        Sorted by value from low to high.

        Returns:
            Tuple[Category, ...]: A tuple of Categories.
        """
        return tuple(sorted(self._categories.values(), key=lambda x: x.value))

    @property
    def coordinates(self) -> Tuple[Coordinate, ...]:
        """Return a tuple of all Coordinates in the RiskMatrix.

        Returns:
            Tuple[Coordinate, ...]: Tuple of Coordinates sorted alphabetically.
        """
        return tuple(sorted(self._coordinates, key=lambda c: str(c)))

    def add_axis(
        self,
        axis_name: str,
        *,
        points: Iterable[Tuple] = None,
        size: int = None,
        use_letters: bool = False,
    ) -> Axis:
        """Add an axis to the risk matrix using a list of axis points.

        Alternatively, you can also give a size number to quickly set up an axis.
        This is nice if you don't care about the information in the axis points.

        Args:
            axis_name (str): The name for the axis. E.g. Severity or Probability.
            points (Iterable[Tuple], optional): A list of points that make up the axis. Defaults to None.
            size (int, optional): A quick way to set up an axis by defining how many points you want. Defaults to None.
            use_letters (bool, optional): Option to use letters instead of numbers when specifying size. Defaults to False.

        Raises:
            ValueError: You have to provide either a list of points or a size. You can't do both.

        Returns:
            Axis
        """
        if points and size:
            raise ValueError(
                "You should choose between giving a list of points or defining a size."
            )

        axis = Axis(axis_name, self)

        if points:
            for point in points:
                p = Point(*point)
                axis.add_point(p)
        elif size:
            for code in range(1, size + 1):
                if use_letters:
                    code = self._convert_number_to_letter(code)
                axis_point = Point(str(code), "")
                axis.add_point(axis_point)

        self.axes.add(axis)

        return axis

    def _convert_number_to_letter(self, number: int):
        """Provide a number between 1 and 26 to return the appropriate letter.

        Args:
            number (int): Number between 1 and 26

        Raises:
            ValueError: If the number is not between 1 and 26.

        Returns:
            str: A single letter equivalent to the number.
        """
        if not 0 <= number <= 26:
            raise ValueError(f"The number {number} has to be between 1 and 26.")

        return "ABCDEFGHIJKLMNOPQRSTUVWXYZ"[number - 1]

    def add_category(
        self, code: str, name: str, color: str, text_color: str, description: str = ""
    ) -> Category:
        """Add a category to the Riskmatrix.

        Categories should be added from low to high.

        Args:
            code (str): A short code for the category. E.g. 'HIG'
            name (str): A full name for the category. E.g. 'High risk'
            color (str): A hexadecimal background color code.
            text_color (str): A hexadecimal text color code.
            description (str, optional): A longer description about what this category means. Defaults to "".

        Returns:
            Category
        """
        category = Category(code, name, color, text_color, description)
        category.value = len(self.categories)
        self._categories[category.value] = category

        return category

    def map_coordinate(
        self, category: Category, points: Iterable[AxisPoint]
    ) -> Coordinate:
        """Map a Category to a Coordinate

        Args:
            category (Category): An instance of Category.
            points (Iterable[AxisPoint]): A collection of AxisPoint that make up a Coordinate.

        Returns:
            Coordinate
        """
        c = Coordinate(points)

        if c.matrix is not self:
            raise ValueError(
                f"This Coordinate {c} does not belong to RiskMatrix {self.name}"
            )

        self._coordinates[c] = self.categories[category.value]
        return c

    def map_coordinates(
        self, category: Category, coordinates: Iterable[Iterable[AxisPoint]]
    ) -> None:
        """Given a Category and a list of AxisPoint collections (each making up a Coordinate), map the Category to
        each Coordinate.

        Args:
            category (Category): A single Category.
            coordinates (Iterable[Iterable[AxisPoint]]): A list of AxisPoint iterables that represent Coordinates.

        Returns:
            None
        """
        for coordinate_points in coordinates:
            self.map_coordinate(category, coordinate_points)

    def get_category(self, coordinate: Coordinate) -> Category:
        """Give a Coordinate to get a Category if there is a mapping between them.

        Args:
            coordinate (Coordinate): An instance of Coordinate.

        Returns:
            Category: An instance of Category.

        Exceptions:
            IndexError: If the Coordinate couldn't be found, an IndexError is raised.
        """
        try:
            return self._coordinates[coordinate]
        except KeyError as e:
            raise KeyError(
                f"{coordinate} couldn't be found. Are you sure you mapped it?"
            ) from e

    def get_coordinate(self, coordinate: str) -> Optional[Coordinate]:
        """Get the Coordinate for a string code like 'A2'.

        Args:
            coordinate (str): A string which is the code of the Coordinate. E.g. 'A2'

        Returns:
            Optional[Coordinate]: A Coordinate if it can be found, or None.
        """
        for c in self._coordinates:
            if str(c) == coordinate:
                return c
        return None

    def get_max_category(self) -> Optional[Category]:
        """Get the Category with the highest value in the RiskMatrix.

        Returns:
            Optional[Category]: The Category with the highest value.
            Will return None if there are no Categories defined in the RiskMatrix.
        """
        return max(self.categories, default=None)



from __future__ import annotations
from collections import namedtuple
from functools import total_ordering
from typing import List, Optional, Tuple, Union

# This is a hack to make mypy happy
if False:
    from .matrix import RiskMatrix

Point = namedtuple("Point", "code name description", defaults=[""])


@total_ordering
class AxisPoint:
    """Has:
    * a code (e.g. 'A' or 1)
    * a name (e.g. 'Unlikely')
    * a description (e.g. For a definition.)
    * a numeric value
    """

    def __init__(
        self, code: str, name: str, description: str, value: int, axis: Axis
    ) -> None:
        self.code = code
        self.name = name
        self.desc = description
        self.value = value
        self.axis = axis

    def __repr__(self):
        return f"AxisPoint{self.code, self.name, self.desc}"

    def __str__(self):
        return f"Point: {self.code} - {self.name}"

    def __eq__(self, other: object) -> bool:
        """Allow equality based on code string or value."""
        if not isinstance(other, (AxisPoint, str)):
            return NotImplemented

        if isinstance(other, str):
            return self.code == other

        return self.value == other.value

    def __lt__(self, other: object) -> bool:
        if not isinstance(other, AxisPoint):
            return NotImplemented

        return self.value < other.value

    def __hash__(self) -> int:
        return hash(f"{self.code} {self.name}")


class Axis:
    """An axis for a RiskMatrix. Contains AxisPoints.

    This class holds the points together and gives them an order.
    """

    def __init__(self, name: str, matrix: RiskMatrix) -> None:
        self.name = name
        self._points: List[AxisPoint] = []
        self.matrix = matrix

    def __repr__(self):
        return f"Axis({self.name})"

    def __str__(self):
        return self.name

    def __getitem__(self, key: int) -> AxisPoint:
        return self.points[key]

    def __len__(self) -> int:
        return len(self.points)

    @property
    def points(self) -> Tuple[AxisPoint, ...]:
        """Get the points of the Axis.

        Returns:
            Tuple[AxisPoint, ...]: An ordered tuple of AxisPoint.
        """
        return tuple(self._points)

    def add_point(self, point: Point) -> None:
        """Add an AxisPoint to the Axis.

        If no value is set, the AxisPoint value is set to the length+1 of the Axis.
        This means it's possible to have a point with a set value, and one without
        that gets set to the same value. In these cases, a ValueError is thrown.

        Args:
            point (Point): The point to add to the Axis.

        Raises:
            ValueError: Is raised when attempting to add an AxisPoint with a value that already exists.

        Returns:
            None
        """
        code, name, description = point
        value = len(self._points) + 1
        axis = self

        axis_point = AxisPoint(code, name, description, value, axis)
        self._points.append(axis_point)
        self._points = sorted(self.points)
